
/* Chrysalide - Outil d'analyse de fichiers binaires
 * instruction.c - gestion des instructions de la ARM
 *
 * Copyright (C) 2013-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Chrysalide.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "instruction.h"


#include <assert.h>
#include <malloc.h>
#include <string.h>


#include <common/extstr.h>


#include "instruction-int.h"



/* Initialise la classe des instructions ARM. */
static void g_arm_instruction_class_init(GArmInstructionClass *);

/* Initialise une instance de instruction ARM. */
static void g_arm_instruction_init(GArmInstruction *);

/* Supprime toutes les références externes. */
static void g_arm_instruction_dispose(GArmInstruction *);

/* Procède à la libération totale de la mémoire. */
static void g_arm_instruction_finalize(GArmInstruction *);

/* Fournit le nom humain de l'instruction manipulée. */
static const char *g_arm_instruction_get_keyword(GArmInstruction *, AsmSyntax);



/* Indique le type défini pour une représentation d'une instruction ARM. */
G_DEFINE_TYPE(GArmInstruction, g_arm_instruction, G_TYPE_ARCH_INSTRUCTION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des instructions ARM.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_class_init(GArmInstructionClass *klass)
{
    GObjectClass *object_class;             /* Autre version de la classe  */
    GArchInstructionClass *instr;           /* Encore une autre vision...  */

    object_class = G_OBJECT_CLASS(klass);
    instr = G_ARCH_INSTRUCTION_CLASS(klass);

    object_class->dispose = (GObjectFinalizeFunc/* ! */)g_arm_instruction_dispose;
    object_class->finalize = (GObjectFinalizeFunc)g_arm_instruction_finalize;

    instr->get_keyword = (get_instruction_keyword_fc)g_arm_instruction_get_keyword;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance à initialiser.                              *
*                                                                             *
*  Description : Initialise une instance d'instruction ARM.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_init(GArmInstruction *instr)
{
    instr->cond = ACC_AL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_dispose(GArmInstruction *instr)
{
    G_OBJECT_CLASS(g_arm_instruction_parent_class)->dispose(G_OBJECT(instr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_finalize(GArmInstruction *instr)
{
    if (instr->suffix != NULL)
        free(instr->suffix);

    if (instr->cached_keyword != NULL)
        free(instr->cached_keyword);

    G_OBJECT_CLASS(g_arm_instruction_parent_class)->finalize(G_OBJECT(instr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr  = instruction quelconque à modifier.                  *
*                suffix = chaîne de caractères fournie en complément.         *
*                                                                             *
*  Description : Etend la désignation d'un nom d'instruction.                 *
*                                                                             *
*  Retour      : true.                                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_arm_instruction_extend_keyword(GArmInstruction *instr, const char *suffix)
{
    instr->suffix = stradd(instr->suffix, suffix);

    if (instr->cached_keyword != NULL)
    {
        free(instr->cached_keyword);
        instr->cached_keyword = NULL;
    }

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr  = instruction d'assemblage à consulter.               *
*                syntax = type de représentation demandée.                    *
*                                                                             *
*  Description : Fournit le nom humain de l'instruction manipulée.            *
*                                                                             *
*  Retour      : Mot clef de bas niveau.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static const char *g_arm_instruction_get_keyword(GArmInstruction *instr, AsmSyntax syntax)
{
    if (instr->cached_keyword == NULL)
    {
        instr->cached_keyword = strdup(instr->keyword);

        if (instr->suffix != NULL)
            instr->cached_keyword = stradd(instr->cached_keyword, instr->suffix);

    }

    return instr->cached_keyword;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instruction ARM à mettre à jour.                     *
*                cond  = condition d'exécution de l'instruction.              *
*                                                                             *
*  Description : Définit les conditions d'exécution d'une instruction ARM.    *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_arm_instruction_set_cond(GArmInstruction *instr, ArmCondCode cond)
{
    bool result;                            /* Bilan à retourner           */
    const char *suffix;                     /* Eventuelle marque à ajouter */

    instr->cond = cond;

    switch (cond)
    {
        case ACC_EQ: suffix = "eq"; break;
        case ACC_NE: suffix = "ne"; break;
        case ACC_HS: suffix = "hs"; break;
        case ACC_LO: suffix = "lo"; break;
        case ACC_MI: suffix = "mi"; break;
        case ACC_PL: suffix = "pl"; break;
        case ACC_VS: suffix = "vs"; break;
        case ACC_VC: suffix = "vc"; break;
        case ACC_HI: suffix = "hi"; break;
        case ACC_LS: suffix = "ls"; break;
        case ACC_GE: suffix = "ge"; break;
        case ACC_LT: suffix = "lt"; break;
        case ACC_GT: suffix = "gt"; break;
        case ACC_LE: suffix = "le"; break;
        case ACC_AL: suffix = NULL; break;
        case ACC_NV: suffix = "nv"; break;

        default:    /* Pour GCC... */
            assert(false);
            suffix = NULL;
            break;

    }

    if (suffix != NULL)
        result = g_arm_instruction_extend_keyword(instr, suffix);

    else
        result = true;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instruction ARM à consulter.                         *
*                                                                             *
*  Description : Indique les conditions d'exécution d'une instruction ARM.    *
*                                                                             *
*  Retour      : Condition d'exécution de l'instruction.                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

ArmCondCode g_arm_instruction_get_cond(const GArmInstruction *instr)
{
    return instr->cond;

}
